package com.sourcebits.flickrhack;

import java.io.IOException;
import java.io.InputStream;
import java.lang.ref.SoftReference;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.Collections;
import java.util.HashMap;

import java.util.Map;
import java.util.WeakHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import android.graphics.drawable.Drawable;
import android.os.Handler;
import android.os.Message;
import android.util.Log;
import android.widget.ImageView;

/**
 * 
 * 
 *         Working with drawables is faster that with bitmaps so I uses
 *         drawables instead
 * 
 *         Using SoftReference is great, but it makes the cached image to be
 *         deleted too often, so I added a Linked list that holds images
 *         references, preventing from the image to be deleted, until it reached
 *         a predefined size
 * 
 *         To open the InputStream I used java.net.URLConnection which allows me
 *         to use web cache (you need to set a response cache first, but that's
 *         another story)
 * 
 */

public class ImageLoader {

	// private final LinkedList<Drawable> mChacheController = new
	// LinkedList<Drawable>();
	private ExecutorService mThreadPool;

	/**
	 * Map to store the URL and the downloaded image (BUFFER)
	 */

	private final Map<String, SoftReference<Drawable>> mCache = new HashMap<String, SoftReference<Drawable>>();

	/**
	 * Map to sore the Image View and the corresponding URL of that image
	 * 
	 * 
	 * Collections.synchronizedMap() guarantees that each atomic operation you
	 * want to run on the map will be synchronized.
	 * 
	 * 
	 */
	private final Map<ImageView, String> mImageViewsMap = Collections
			.synchronizedMap(new WeakHashMap<ImageView, String>());

	public static int MAX_CACHE_SIZE = 200;
	public int THREAD_POOL_SIZE = 3;
	private Drawable drawable; // the image

	/**
	 * Constructor
	 * 
	 * 
	 * public static ExecutorService newFixedThreadPool(int nThreads) Creates a
	 * thread pool that reuses a fixed number of threads operating off a shared
	 * unbounded queue. At any point, at most nThreads threads will be active
	 * processing tasks. If additional tasks are submitted when all threads are
	 * active, they will wait in the queue until a thread is available. If any
	 * thread terminates due to a failure during execution prior to shutdown, a
	 * new one will take its place if needed to execute subsequent tasks. The
	 * threads in the pool will exist until it is explicitly shutdown.
	 * Parameters: nThreads - the number of threads in the pool Returns: the
	 * newly created thread pool Throws: IllegalArgumentException - if nThreads
	 * <= 0
	 */
	public ImageLoader() {
		mThreadPool = Executors.newFixedThreadPool(THREAD_POOL_SIZE);
	}

	/**
	 * Clears all instance data and stops running threads
	 */
	public void Reset() {
		ExecutorService oldThreadPool = mThreadPool;
		mThreadPool = Executors.newFixedThreadPool(THREAD_POOL_SIZE);
		oldThreadPool.shutdownNow();

		// mChacheController.clear();
		mCache.clear();
		mImageViewsMap.clear();
	}

	public void loadDrawable(final String url, final ImageView imageView,
			Drawable placeholder) {

		mImageViewsMap.put(imageView, url); // put the imageview and its url in
		// Map

		// get the image from the cache (buffer)
		drawable = getDrawableFromCache(url);

		// check in UI thread, so no concurrency issues
		// if the image is in the cache display it
		if (drawable != null) {
			// Log.d(null, "Item loaded from mCache: " + url);
			imageView.setImageDrawable(drawable);
		} else { // display the dummy image & queue job
			imageView.setImageDrawable(placeholder);
			queueJob(url, imageView, placeholder);
		}
	}

	private Drawable getDrawableFromCache(String url) {

		// if the image is in the cache then return it
		if (mCache.containsKey(url)) {
			return mCache.get(url).get();
		}

		// image not found return null
		return null;
	}

	private synchronized void putDrawableInCache(String url, Drawable drawable) {
		// int chacheControllerSize = mChacheController.size();
		// if (chacheControllerSize > MAX_CACHE_SIZE)
		// mChacheController.subList(0, MAX_CACHE_SIZE / 2).clear();

		// mChacheController.addLast(drawable);
		mCache.put(url, new SoftReference<Drawable>(drawable));

	}

	private void queueJob(final String url, final ImageView imageView,
			final Drawable placeholder) {
		/* Create handler in UI thread. */

		final Handler handler = new Handler() {

			@Override
			public void handleMessage(Message msg) {
				String tag = mImageViewsMap.get(imageView);
				if (tag != null && tag.equals(url)) {
					if (imageView.isShown())
						if (msg.obj != null) {
							imageView.setImageDrawable((Drawable) msg.obj);
						} else {
							imageView.setImageDrawable(placeholder);
							// Log.d(null, "fail " + url);
						}
				}
			}
		};

		mThreadPool.submit(new Runnable() {
			public void run() {
				final Drawable bmp = downloadDrawable(url);
				// if the view is not visible anymore, the image will be ready
				// for next time in cache
				if (imageView.isShown()) {
					Message message = Message.obtain(); // equvalent to new
														// Message()
					message.obj = bmp;
					// Log.d(null, "Item downloaded: " + url);

					handler.sendMessage(message);
				}
			}
		});
	}

	// Download the image from the specified URL
	private Drawable downloadDrawable(String url) {
		try {
			InputStream is = getInputStream(url);

			Drawable drawable = Drawable.createFromStream(is, url); // actual
																	// download
			putDrawableInCache(url, drawable);
			return drawable;

		} catch (Exception e) {

			Log.v("URL ", "==" + url);
			Log.v("EXCEPTION in downloadDrawable()==", "" + e.toString());
		}

		return null;
	}

	private InputStream getInputStream(String urlString)
			throws MalformedURLException, IOException {
		URL url = new URL(urlString);
		URLConnection connection;
		connection = url.openConnection();
		connection.setUseCaches(true);
		connection.connect();
		InputStream response = connection.getInputStream();

		return response;
	}

	/**
	 * If you move out of the UI, then call this method
	 */
	public void KillTheProcess() {
		Log.v("ImageLoader.java", "=About to be killed");
		// mThreadPool.shutdownNow();
		mThreadPool.shutdown();
		Log.v("ImageLoader.java", "=KILLED");
	}
}